from typing import OrderedDict


def weekdays_bitmask(data: OrderedDict[str, bool]) -> str:
    """Convert ordered dict with weekdays to a bitmask.

    ### Args:
        * data (OrderedDict[str, bool]): OrderedDict formatted as follows: OrderedDict(Monday=bool, Tuesday=bool, Wednesday=bool, Thursday=bool, Friday=bool, Saturday=bool, Sunday=bool)

    ### Returns:
        * str: _description_
    """
    if len(data) != 7:
        raise ValueError(
            "OrderedDict must be formatted as follows: OrderedDict(Monday=bool, Tuesday=bool, Wednesday=bool, Thursday=bool, Friday=bool, Saturday=bool, Sunday=bool)"
        )

    return "".join("1" if data[day] else "0" for day in data)



from enum import Enum, auto


class PlatformTypeType(Enum):
    """Enumeration used to declare types of platform type.

    * U - Undefined
    * PL - Platform/track at train station
    * ST - Stop at bus or tram station
    * GA - Terminal/Gate at airport
    * PI - Pier if ship or ferry
    * SL - Slot/parking space if bike or car
    * FL - Floor in buildings or at footpath
    * CI - Check-in/entrance
    * CO - Check-out/exit
    * X - No explicit type
    * H - Hide platform information
    """

    U = auto()
    "Undefined"

    PL = auto()
    "Platform/track at train station"

    ST = auto()
    "Stop at bus or tram station"

    GA = auto()
    "Terminal/Gate at airport"

    PI = auto()
    "Pier if ship or ferry"

    SL = auto()
    "Slot/parking space if bike or car"

    FL = auto()
    "Floor in buildings or at footpath"

    CI = auto()
    "Check-in/entrance"

    CO = auto()
    "Check-out/exit"

    X = auto()
    "No explicit type"

    H = auto()
    "Hide platform information"



from datetime import datetime
from typing import Any, List, Mapping, Union

from ..classes.journey import Journey
from ..classes.message import Message
from ..classes.stop import Stop
from ..utility import ref_upgrade


class LineArrival:
    def __init__(
        self,
        data: Mapping[str, Any],
        retrieve_stops: bool = False,
        retrieve_journey: bool = False,
        client: Union["pyrmv.Client", None] = None,
    ):
        if (retrieve_stops or retrieve_journey) and client is None:
            raise KeyError(
                "Stops/journey retrieval requested but client argument was not provided."
            )

        # Upgrade is temporarily used due to RMV API mismatch
        # self.journey = client.journey_detail(data["JourneyDetailRef"]["ref"])
        self.journey: Union[Journey, None] = (
            client.journey_detail(ref_upgrade(data["JourneyDetailRef"]["ref"]))
            if retrieve_journey
            else None
        )

        self.status: str = data["JourneyStatus"]
        self.messages: List[Message] = []
        self.name: str = data["name"]
        self.type: str = data["type"]
        self.stop_name: str = data["stop"]
        self.stop_id: str = data["stopid"]
        self.stop_id_ext: Union[str, None] = data.get("stopExtId")
        self.stop: Union[Stop, None] = (
            client.stop_by_id(self.stop_id) if retrieve_stops else None
        )
        self.time: datetime = datetime.strptime(
            f"{data['date']} {data['time']}", "%Y-%m-%d %H:%M:%S"
        )
        self.time_real: datetime = (
            datetime.strptime(f"{data['rtDate']} {data['rtTime']}", "%Y-%m-%d %H:%M:%S")
            if data.get("rtTime") and data.get("rtDate")
            else None
        )
        self.reachable: bool = data["reachable"]
        self.origin: Union[str, None] = data.get("origin")

        if "Messages" in data:
            self.messages.extend(Message(message) for message in data["Messages"]["Message"])

    def __str__(self) -> str:
        return (
            f"{self.name} coming from {self.origin} at {self.time.time()} {self.date.date()}"
        )


class LineDeparture:
    def __init__(
        self,
        data: Mapping[str, Any],
        retrieve_stops: bool = False,
        retrieve_journey: bool = False,
        client: Union["pyrmv.Client", None] = None,
    ):
        if (retrieve_stops or retrieve_journey) and client is None:
            raise KeyError(
                "Stops/journey retrieval requested but client argument was not provided."
            )

        # Upgrade is temporarily used due to RMV API mismatch
        # self.journey = client.journey_detail(data["JourneyDetailRef"]["ref"])
        self.journey: Union[Journey, None] = (
            client.journey_detail(ref_upgrade(data["JourneyDetailRef"]["ref"]))
            if retrieve_journey
            else None
        )

        self.status: str = data["JourneyStatus"]
        self.messages: List[Message] = []
        self.name: str = data["name"]
        self.type: str = data["type"]
        self.stop_name: str = data["stop"]
        self.stop_id: str = data["stopid"]
        self.stop_id_ext: Union[str, None] = data.get("stopExtId")
        self.stop: Union[Stop, None] = (
            client.stop_by_id(self.stop_id) if retrieve_stops else None
        )
        self.time: datetime = datetime.strptime(
            f"{data['date']} {data['time']}", "%Y-%m-%d %H:%M:%S"
        )
        self.time_real: datetime = (
            datetime.strptime(f"{data['rtDate']} {data['rtTime']}", "%Y-%m-%d %H:%M:%S")
            if data.get("rtTime") and data.get("rtDate")
            else None
        )
        self.reachable: bool = data["reachable"]
        self.direction: Union[str, None] = data.get("direction")
        self.direction_flag: Union[str, None] = data.get("directionFlag")

        if "Messages" in data:
            self.messages.extend(Message(message) for message in data["Messages"]["Message"])

    def __str__(self) -> str:
        return (
            f"{self.name} heading {self.direction} at {self.time.time()} {self.date.date()}"
        )


class BoardArrival(list):
    def __init__(
        self,
        data: Mapping[str, Any],
        retrieve_stops: bool = False,
        retrieve_journey: bool = False,
        client: Union["pyrmv.Client", None] = None,
    ):
        """Arrival board representation

        ### Args:
            * data (`Mapping[str, Any]`): Dictionary from RMV to be parsed.
            * retrieve_stops (`bool`, *optional*): Retrieve `Stop` objects for each line of the board. Defaults to `False`.
            * retrieve_journey (`bool`, *optional*): Retrieve `Journey` object for each line of the board. Defaults to `False`.
            * client (`Union[Client, None]`, *optional*): Client to be used if `retrieve_stops` or `retrieve_journey` are set to `True`. Defaults to `None`.
        """
        super().__init__([])

        if "Arrival" not in data:
            return

        for line in data["Arrival"]:
            self.append(
                LineArrival(
                    line,
                    retrieve_stops=retrieve_stops,
                    retrieve_journey=retrieve_journey,
                    client=client,
                )
            )

    def __str__(self) -> str:
        return "Arrival board\n" + "\n".join([str(line) for line in self])


class BoardDeparture(list):
    def __init__(
        self,
        data: Mapping[str, Any],
        retrieve_stops: bool = False,
        retrieve_journey: bool = False,
        client: Union["pyrmv.Client", None] = None,
    ):
        """Departure board representation

        ### Args:
            * data (`Mapping[str, Any]`): Dictionary from RMV to be parsed.
            * retrieve_stops (`bool`, *optional*): Retrieve `Stop` objects for each line of the board. Defaults to `False`.
            * retrieve_journey (`bool`, *optional*): Retrieve `Journey` object for each line of the board. Defaults to `False`.
            * client (`Union[Client, None]`, *optional*): Client to be used if `retrieve_stops` or `retrieve_journey` are set to `True`. Defaults to `None`.
        """
        super().__init__([])

        if "Departure" not in data:
            return

        for line in data["Departure"]:
            self.append(
                LineDeparture(
                    line,
                    retrieve_stops=retrieve_stops,
                    retrieve_journey=retrieve_journey,
                    client=client,
                )
            )

    def __str__(self) -> str:
        return "Departure board\n" + "\n".join([str(line) for line in self])



from typing import Any, List, Mapping, Union

from ..classes.message import Message
from ..classes.stop import Stop
from ..utility import ref_upgrade


class Journey:
    """Journey object."""

    def __init__(self, data: Mapping[str, Any]):
        self.stops: List[Stop] = []

        # Upgrade is temporarily used due to RMV API mismatch
        # self.ref = data["ref"]
        self.ref: str = ref_upgrade(data["ref"])

        self.direction: Union[str, None] = (
            data["Directions"]["Direction"][0].get("value")
            if data["Directions"]["Direction"]
            else None
        )
        self.direction_flag: Union[str, None] = (
            data["Directions"]["Direction"][0].get("flag")
            if data["Directions"]["Direction"]
            else None
        )
        self.stops.extend(Stop(stop) for stop in data["Stops"]["Stop"])
        self.messages: List[Message] = []

        if "Messages" in data:
            self.messages.extend(Message(message) for message in data["Messages"]["Message"])

    def __str__(self) -> str:
        return f"Journey with total of {len(self.stops)} stops and {len(self.messages)} messages heading {self.direction} ({self.direction_flag})"



from enum import auto

from .auto_name import AutoName


class SearchMode(AutoName):
    """Enumeration used to declare types of HIM search modes."""

    MATCH = auto()
    "Iterate over all trips to find HIM messages."

    NOMATCH = auto()
    "Iterate over all HIM messages available."

    TFMATCH = auto()
    "Uses filters defined `metas` parameter."



from datetime import timedelta
from typing import Any, List, Mapping, Union

from isodate import Duration, parse_duration

from ..classes.gis import Gis
from ..classes.message import Message
from ..classes.stop import StopTrip


class Leg:
    """Trip leg object."""

    def __init__(self, data: Mapping[str, Any]):
        self.origin: StopTrip = StopTrip(data["Origin"])
        self.destination: StopTrip = StopTrip(data["Destination"])
        self.gis = (
            None if "GisRef" not in data else Gis(data["GisRef"]["ref"], data["GisRoute"])
        )
        self.messages: List[Message] = []
        self.index: Union[int, None] = data.get("idx")
        self.name: str = data["name"]
        self.type: Union[str, None] = data.get("type")
        self.direction: Union[str, None] = data.get("direction")
        self.number: Union[str, None] = data.get("number")
        self.duration: Union[Duration, timedelta] = parse_duration(data["duration"])
        self.distance: Union[int, None] = data.get("dist")

        if "Messages" in data:
            self.messages.extend(Message(message) for message in data["Messages"]["Message"])



from datetime import datetime, timedelta
from typing import Any, List, Mapping, Union

from isodate import Duration, parse_duration

from ..classes.stop import Stop


class Url:
    """Traffic message channel url object."""

    def __init__(self, data: Mapping[str, Any]) -> None:
        self.name: str = data["name"]
        self.url: str = data["url"]

    def __str__(self) -> str:
        return f"{self.name}: {self.url}"


class Channel:
    """Traffic message channel object."""

    def __init__(self, data: Mapping[str, Any]) -> None:
        self.name: str = data["name"]
        url = []
        url.extend(Url(link) for link in url)
        self.url: List[Url] = url
        self.time_start: Union[datetime, None] = (
            None
            if "validFromDate" not in data
            else datetime.strptime(
                f"{data['validFromDate']} {data['validFromTime']}", "%Y-%m-%d %H:%M:%S"
            )
        )
        self.time_end: Union[datetime, None] = (
            None
            if "validToDate" not in data
            else datetime.strptime(
                f"{data['validToDate']} {data['validToTime']}", "%Y-%m-%d %H:%M:%S"
            )
        )

    def __str__(self) -> str:
        return f"{self.name}: from {self.time_start} until {self.time_end}"


class Message:
    """Traffic message object."""

    def __init__(self, data: Mapping[str, Any]) -> None:
        self.affected_stops: List[Stop] = []
        self.valid_from_stop: Union[Stop, None] = (
            None if "validFromStop" not in data else Stop(data["validFromStop"])
        )
        self.valid_to_stop: Union[Stop, None] = (
            None if "validToStop" not in data else Stop(data["validToStop"])
        )
        self.channels: Union[Channel, None] = []
        self.channels.extend(Channel(channel) for channel in data["channel"])
        self.id: str = data["id"]
        self.active: bool = data["act"]
        self.head: str = "" if "head" not in data else data["head"]
        self.lead: str = "" if "lead" not in data else data["lead"]
        self.text: str = "" if "text" not in data else data["text"]
        self.company: Union[str, None] = data.get("company")
        self.category: Union[str, None] = data.get("category")
        self.priority: Union[int, None] = data.get("priority")
        self.products: int = data["products"]
        self.icon: Mapping[str, Any] = data["icon"]
        self.time_start: Union[datetime, None] = (
            None
            if "validFromDate" not in data or "validFromTime" not in data
            else datetime.strptime(
                f"{data['validFromDate']} {data['validFromTime']}", "%Y-%m-%d %H:%M:%S"
            )
        )
        self.time_end: Union[datetime, None] = (
            None
            if "validToDate" not in data or "validToTime" not in data
            else datetime.strptime(
                f"{data['validToDate']} {data['validToTime']}", "%Y-%m-%d %H:%M:%S"
            )
        )
        self.date_start_alt: Union[str, None] = (
            None if "altStart" not in data else data["altStart"]
        )
        self.date_end_alt: Union[str, None] = None if "altEnd" not in data else data["altEnd"]
        self.time_modified: Union[datetime, None] = (
            None
            if "modDate" not in data or "modTime" not in data
            else datetime.strptime(
                f"{data['modDate']} {data['modTime']}", "%Y-%m-%d %H:%M:%S"
            )
        )
        self.daily_start: Union[datetime, None] = (
            None
            if "dailyStartingAt" not in data
            else datetime.strptime(data["dailyStartingAt"], "%H:%M:%S")
        )
        self.daily_duration: Union[Duration, timedelta] = parse_duration(
            data["dailyDuration"]
        )
        self.base_type: Union[str, None] = data.get("baseType")

        if "affectedStops" in data:
            self.affected_stops.extend(
                Stop(stop) for stop in data["affectedStops"]["StopLocation"]
            )

    def __str__(self) -> str:
        return f"{self.base_type} message with priority {self.products} valid from {self.time_start} until {self.time_end}: {self.head} - {self.lead}"



from datetime import datetime, timedelta
from typing import List

import pytest

from pyrmv import Client, enums
from pyrmv.classes import BoardArrival, BoardDeparture, Journey, Message, Stop, Trip


def test_board_arrival(api_client: Client, sample_stop_id: str):
    assert isinstance(
        api_client.board_arrival(id=sample_stop_id, journeys_max=3), BoardArrival
    )


def test_board_departure(api_client: Client, sample_stop_id: str):
    assert isinstance(
        api_client.board_departure(id=sample_stop_id, journeys_max=3), BoardDeparture
    )


def test_him_search(api_client: Client):
    response = api_client.him_search(time_end=datetime.now() + timedelta(days=10))

    if len(response) != 0:
        assert isinstance(response[0], Message)
    else:
        assert isinstance(response, list)


def test_journey_detail(api_client: Client, sample_journey_id: str):
    assert isinstance(
        api_client.journey_detail(
            sample_journey_id,
            real_time_mode=enums.RealTimeMode.FULL,
        ),
        Journey,
    )


def test_stop_by_coords(api_client: Client, sample_origin: List[str]):
    assert isinstance(
        api_client.stop_by_coords(sample_origin[0], sample_origin[1], max_number=3)[0], Stop
    )


def test_stop_by_id(api_client: Client, sample_stop_id: str):
    assert isinstance(api_client.stop_by_id(sample_stop_id), Stop)


def test_trip_find(
    api_client: Client, sample_origin: List[str], sample_destination: List[float]
):
    assert isinstance(
        api_client.trip_find(
            origin_coord_lat=sample_origin[0],
            origin_coord_lon=sample_origin[1],
            destination_coord_lat=sample_destination[0],
            destination_coord_lon=sample_destination[1],
            messages=True,
        )[0],
        Trip,
    )


def test_trip_recon(
    api_client: Client, sample_origin: List[str], sample_destination: List[float]
):
    assert isinstance(
        api_client.trip_recon(
            api_client.trip_find(
                origin_coord_lat=sample_origin[0],
                origin_coord_lon=sample_origin[1],
                destination_coord_lat=sample_destination[0],
                destination_coord_lon=sample_destination[1],
                messages=True,
            )[0],
        )[0],
        Trip,
    )


def test_stop_by_name(api_client: Client):
    assert isinstance(api_client.stop_by_name("Hauptwache", max_number=1)[0], Stop)



from .board import BoardArrival, BoardDeparture, LineArrival, LineDeparture
from .gis import Gis
from .journey import Journey
from .leg import Leg
from .message import Channel, Message, Url
from .platform_type import PlatformType
from .stop import Stop, StopTrip
from .ticket import Ticket
from .trip import Trip



from datetime import datetime
from typing import Union


class Stop:
    """Stop object."""

    def __init__(self, data: dict):
        self.name: str = data["name"]
        self.id: str = data["id"]
        self.ext_id: Union[str, None] = data.get("extId")
        self.description: Union[str, None] = data.get("description")
        self.lon: float = data["lon"]
        self.lat: float = data["lat"]
        self.route_index: Union[int, None] = data.get("routeIdx")
        self.track_arrival: Union[str, None] = data.get("arrTrack")
        self.track_departure: Union[str, None] = data.get("depTrack")

    def __str__(self) -> str:
        return f"Stop {self.name} at {self.lon}, {self.lat}"


class StopTrip(Stop):
    """Trip stop object. It's like a Stop object, but with a date and time."""

    def __init__(self, data: dict):
        self.type: str = data["type"]
        self.time: datetime = datetime.strptime(
            f"{data['date']} {data['time']}", "%Y-%m-%d %H:%M:%S"
        )
        super().__init__(data)

    def __str__(self) -> str:
        return f"Stop {self.name} at {self.lon}, {self.lat} at {self.time.time()}"



from enum import auto

from .auto_name import AutoName


class AffectedJourneyStopMode(AutoName):
    """Enumeration used to declare types of affected journey stops return modes."""

    ALL = auto()
    "Return all affected stops of affected journeys."

    IMP = auto()
    "Return important stops of affected journeys."

    OFF = auto()
    "Do not return stops of affected journeys."



from typing import Any, Mapping, Union

from ..enums.platform_type_type import PlatformTypeType


class PlatformType:
    """Platform information."""

    def __init__(self, data: Mapping[str, Any]):
        self.type: PlatformTypeType = (
            PlatformTypeType.U if "type" not in data else PlatformTypeType(data.get("type"))
        )
        self.text: Union[str, None] = data.get("text")
        self.hidden: bool = bool(data.get("hidden"))
        self.lon: float = data["lon"]
        self.lat: float = data["lat"]
        self.alt: int = data["alt"]



from enum import auto

from .auto_name import AutoName


class Language(AutoName):
    """Enumeration used to declare locales as ISO-3166 codes (but only available in HAFAS ones)."""

    AR = auto()
    "Arabic"

    CA = auto()
    "Catalan, Valencian"

    DA = auto()
    "Danish"

    DE = auto()
    "German"

    EL = auto()
    "Greek"

    EN = auto()
    "English"

    ES = auto()
    "Spanish"

    FI = auto()
    "Finnish"

    FR = auto()
    "French"

    HI = auto()
    "Hindi"

    HR = auto()
    "Croatian"

    HU = auto()
    "Hungarian"

    IT = auto()
    "Italian"

    NL = auto()
    "Dutch"

    NO = auto()
    "Norwegian"

    PL = auto()
    "Polish"

    RU = auto()
    "Russian"

    SK = auto()
    "Slovak"

    SL = auto()
    "Slovenian"

    SV = auto()
    "Swedish"

    TL = auto()
    "Tagalog"

    TR = auto()
    "Turkish"

    UR = auto()
    "Urdu"

    ZH = auto()
    "Chinese"



def ref_upgrade(ref: str) -> str:
    """This function converts older journey refs to the newer ones.

    ### WARNING
    This function will be deprecated as soon as RMV updates their API

    ### Args:
        * ref (`str`): Old ref like this one: `2|#VN#1#ST#1700765441#PI#0#ZI#160749#TA#0#DA#241123#1S#3004646#1T#2228#LS#3006907#LT#2354#PU#80#RT#1#CA#S30#ZE#S1#ZB#      S1#PC#3#FR#3004646#FT#2228#TO#3006907#TT#2354#`

    ### Raises:
        * `KeyError`: Some required keys are not found in the ref provided

    ### Returns:
        * `str`: Ref of the new type
    """

    items = "|".join(ref.split("|")[1:]).strip("#").split("#")
    result = {items[i]: items[i + 1] for i in range(0, len(items), 2)}

    for required in ["VN", "ZI", "TA", "PU"]:
        if required not in result:
            raise KeyError(
                f"Required key {required} in the old journey ref is not found during conversion to the newer journey ref"
            )

    return "|".join([result["VN"], result["ZI"], result["TA"], result["PU"]])



from datetime import timedelta
from typing import Any, Mapping, Union

from isodate import Duration, parse_duration


class Gis:
    """Gis object."""

    def __init__(self, ref: str, route: Mapping[str, Any]):
        self.ref: str = ref
        self.dist: Union[int, None] = route.get("dist")
        self.duration: Union[Duration, timedelta] = parse_duration(route["durS"])
        self.geo: Union[int, None] = route.get("dirGeo")



from datetime import timedelta
from typing import List, Union

from isodate import Duration, parse_duration

from ..classes.leg import Leg
from ..classes.stop import StopTrip


class Trip:
    """Trip object."""

    def __init__(self, data: dict):
        self.origin: StopTrip = StopTrip(data["Origin"])
        self.destination: StopTrip = StopTrip(data["Destination"])
        self.legs: List[Leg] = []
        self.legs.extend(Leg(leg) for leg in data["LegList"]["Leg"])
        self.calculation: str = data["calculation"]
        self.index: int = data["idx"]
        self.id: str = data["tripId"]
        self.ctx_recon: str = data["ctxRecon"]
        self.duration: Union[Duration, timedelta, None] = (
            None if "duration" not in data else parse_duration(data["duration"])
        )
        self.real_time_duration: Union[Duration, timedelta, None] = (
            None if "rtDuration" not in data else parse_duration(data["rtDuration"])
        )
        self.checksum: str = data["checksum"]
        self.transfer_count: int = data.get("transferCount", 0)

    def __str__(self) -> str:
        return f"Trip from {self.origin.name} to {self.destination.name} lasting {self.duration} ({self.real_time_duration}) with {len(self.legs)} legs and {self.transfer_count} transfers"
